/***************************************************************************
 *   Copyright (C) 2008 by Volker Lanz <vl@fidra.de>                       *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA            *
 ***************************************************************************/

#include "core/libparted.h"
#include "core/device.h"
#include "core/partition.h"
#include "core/partitiontable.h"
#include "core/operationstack.h"

#include "jobs/setpartflagsjob.h"

#include "fs/filesystem.h"
#include "fs/filesystemfactory.h"

#include "util/globallog.h"

#include <QString>
#include <QStringList>
#include <QList>
#include <QFile>
#include <QMap>

#include <kdebug.h>
#include <klocale.h>

#include <parted/parted.h>
#include <sys/statvfs.h>
#include <unistd.h>

static PedExceptionOption pedExceptionHandler(PedException* e)
{
	log(log::error) << i18nc("@info/plain", "LibParted Exception: %1", QString::fromLocal8Bit(e->message));
	return PED_EXCEPTION_UNHANDLED;
}

static QString findUuidDevice(const QString& s)
{
	const QString filename = "/dev/disk/by-uuid/" + QString(s).remove("UUID=", Qt::CaseInsensitive);
	return QFile::exists(filename) ? QFile::symLinkTarget(filename) : "";
}

static void readMountpoints(const QString& filename, QMap<QString, QStringList>& result)
{
	QFile file(filename);
	if (!file.open(QIODevice::ReadOnly))
		return;

	QByteArray line = file.readLine();

	while(!line.isEmpty())
	{
		line = line.simplified();

		if (!line.isEmpty() && (line[0] == '/' || line.startsWith("UUID=")))
		{
			QList<QByteArray> split = line.split(' ');

			if (split.size() >= 2)
			{
				QString device = split[0];

				if (device.startsWith("UUID=", Qt::CaseInsensitive))
					device = findUuidDevice(device);

				if (!device.isEmpty())
				{
					QString mountPoint = split[1].replace("\\040", " ");

					if (QFile::exists(mountPoint) && result[device].indexOf(mountPoint) == -1)
						result[device].append(mountPoint);
				}
			}
		}

		line = file.readLine();
	}
}

static qint64 readSectorsUsedLibParted(PedDisk* pedDisk, const Partition& p)
{
	Q_ASSERT(pedDisk);

	qint64 rval = -1;
	
	PedPartition* pedPartition = ped_disk_get_partition_by_sector(pedDisk, p.firstSector());

	if (pedPartition)
	{
		PedFileSystem* pedFileSystem = ped_file_system_open(&pedPartition->geom);

		if (pedFileSystem)
		{
			if (PedConstraint* pedConstraint = ped_file_system_get_resize_constraint(pedFileSystem))
			{
				rval = pedConstraint->min_size;
				ped_constraint_destroy(pedConstraint);
			}

			ped_file_system_close(pedFileSystem);
		}
	}

	return rval;
}

static void readSectorsUsed(PedDisk* pedDisk, Partition& p, QMap<QString, QStringList>& mountInfo)
{
	Q_ASSERT(pedDisk);

	struct statvfs sfs;

	if (p.isMounted() && mountInfo[p.deviceNode()].size() > 0 && statvfs(mountInfo[p.deviceNode()][0].toLatin1(), &sfs) == 0)
		p.fileSystem().setSectorsUsed((sfs.f_blocks - sfs.f_bfree) * sfs.f_bsize / p.sectorSize());
	else if (p.fileSystem().supportGetUsed() == FileSystem::SupportExternal)
		p.fileSystem().setSectorsUsed(p.fileSystem().readUsedCapacity(p.deviceNode()) / p.sectorSize());
	else if (p.fileSystem().supportGetUsed() == FileSystem::SupportLibParted)
		p.fileSystem().setSectorsUsed(readSectorsUsedLibParted(pedDisk, p));
}

static void scanDevicePartitions(PedDevice* pedDevice, Device& d, PedDisk* pedDisk, QMap<QString, QStringList>& mountInfo)
{
	Q_ASSERT(pedDevice);
	Q_ASSERT(pedDisk);

	PedPartition* pedPartition = NULL;
	
	while ((pedPartition = ped_disk_next_partition(pedDisk, pedPartition)))
	{
		if (pedPartition->geom.end - pedPartition->geom.start < Partition::minimumPartitionSectors() || pedPartition->num < 1)
			continue;

		PartitionRole::Roles r = PartitionRole::None;
		FileSystem::Type type = Job::detectFileSystem(pedDevice, pedPartition);

		switch(pedPartition->type)
		{
			case PED_PARTITION_NORMAL:
				r = PartitionRole::Primary;
				break;
				
			case PED_PARTITION_EXTENDED:
				r = PartitionRole::Extended;
				type = FileSystem::Extended;
				break;
				
			case PED_PARTITION_LOGICAL:
				r = PartitionRole::Logical;
				break;
			
			default:
				continue;
		}

		// Find an extended partition this partition is in.
		PartitionNode* parent = d.partitionTable().findPartitionBySector(pedPartition->geom.start, PartitionRole(PartitionRole::Extended));

		// None found, so it's a primary in the device's partition table.
		if (parent == NULL)
			parent = &d.partitionTable();

		const QString node = pedDisk->dev->path + QString::number(pedPartition->num);
		FileSystem* fs = FileSystemFactory::create(type, pedPartition->geom.start, pedPartition->geom.end);

		Partition* part = new Partition(parent, d, PartitionRole(r), fs, pedPartition->geom.start, pedPartition->geom.end,
				pedPartition->num, SetPartFlagsJob::availableFlags(pedPartition),
				mountInfo[node], ped_partition_is_busy(pedPartition), SetPartFlagsJob::activeFlags(pedPartition));

		readSectorsUsed(pedDisk, *part, mountInfo);

		if (fs->supportGetLabel() == FileSystem::SupportExternal)
			fs->setLabel(fs->readLabel(part->deviceNode()));

		parent->append(part);

		PartitionTable::isSnapped(d, *part);
	}

	d.partitionTable().updateUnallocated(d);
	
	ped_disk_destroy(pedDisk);
}

LibParted::LibParted()
{
	ped_exception_set_handler(pedExceptionHandler);
}

void LibParted::scanDevices(OperationStack& ostack)
{
	QMap<QString, QStringList> mountInfo;
	
	readMountpoints("/proc/mounts", mountInfo);
	readMountpoints("/etc/mtab", mountInfo);
	readMountpoints("/etc/fstab", mountInfo);

	ostack.clearOperations();
	ostack.clearDevices();

	ped_device_probe_all();

	PedDevice* pedDevice = ped_device_get_next(NULL);

	while (pedDevice)
	{
		log(log::information) << i18nc("@info/plain", "Device found: %1", pedDevice->model);

		Device* d = new Device(pedDevice->model, pedDevice->path, pedDevice->bios_geom.heads, pedDevice->bios_geom.sectors, pedDevice->bios_geom.cylinders, pedDevice->sector_size);

		PedDisk* pedDisk = ped_disk_new(pedDevice);

		if (pedDisk)
		{
			d->partitionTable().setMaxPrimaries(ped_disk_get_max_primary_partition_count(pedDisk));
			d->partitionTable().setTypeName(pedDisk->type->name);

			scanDevicePartitions(pedDevice, *d, pedDisk, mountInfo);
		}

		ostack.addDevice(d);

		pedDevice = ped_device_get_next(pedDevice);
	}
}
