/***************************************************************************
 *   Copyright (C) 2008 by Volker Lanz <vl@fidra.de>                       *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA            *
 ***************************************************************************/

#include "core/partitionnode.h"

#include "core/partition.h"
#include "core/partitionrole.h"

#include "fs/filesystem.h"

#include <QList>

#include <kdebug.h>

const Partition* PartitionNode::predecessor(const Partition& p) const
{
	Q_ASSERT(p.parent());

	const Partitions& plist = p.parent()->isRoot() == false ? p.parent()->children() : children();

	for (int idx = 1; idx < plist.size(); idx++)
		if (plist[idx] == &p)
			return plist[idx - 1];

	return NULL;
}

const Partition* PartitionNode::successor(const Partition& p) const
{
	Q_ASSERT(p.parent());

	const Partitions& plist = p.parent()->isRoot() == false ? p.parent()->children() : children();

	for (int idx = plist.size() - 2; idx >= 0; idx--)
		if (plist[idx] == &p)
			return plist[idx + 1];

	return NULL;
}

Partition* PartitionNode::predecessor(Partition& p)
{
	Q_ASSERT(p.parent());

	Partitions& plist = p.parent()->isRoot() == false ? p.parent()->children() : children();

	for (int idx = 1; idx < plist.size(); idx++)
		if (plist[idx] == &p)
			return plist[idx - 1];

	return NULL;
}

Partition* PartitionNode::successor(Partition& p)
{
	Q_ASSERT(p.parent());

	Partitions& plist = p.parent()->isRoot() == false ? p.parent()->children() : children();

	for (int idx = plist.size() - 2; idx >= 0; idx--)
		if (plist[idx] == &p)
			return plist[idx + 1];

	return NULL;
}

bool PartitionNode::insert(Partition* p)
{
	if (p == NULL)
		return false;

	for (int idx = 0; idx < children().size(); idx++)
	{
		if (children()[idx]->firstSector() > p->firstSector())
		{
			children().insert(idx, p);
			return true;
		}
	}

	children().insert(children().size(), p);

	return true;
}

bool PartitionNode::remove(Partition* p)
{
	if (p == NULL)
		return false;

	if (children().removeOne(p))
		return true;

	return false;
}

void PartitionNode::clearChildren()
{
	qDeleteAll(children());
	children().clear();
}

Partition* PartitionNode::findPartitionBySector(qint64 s, const PartitionRole& role)
{
	foreach (Partition* p, children())
	{
		// (women and) children first. ;-)
		foreach (Partition* child, p->children())
			if ((child->roles().roles() & role.roles()) && s >= child->firstSector() && s <= child->lastSector())
				return child;
		
		if ((p->roles().roles() & role.roles()) && s >= p->firstSector() && s <= p->lastSector())
			return p;
	}

	return NULL;
}

const Partition* PartitionNode::findPartitionBySector(qint64 s, const PartitionRole& role) const
{
	foreach (const Partition* p, children())
	{
		foreach (const Partition* child, p->children())
			if ((child->roles().roles() & role.roles()) && s >= child->firstSector() && s <= child->lastSector())
				return child;

		if ((p->roles().roles() & role.roles()) && s >= p->firstSector() && s <= p->lastSector())
			return p;
	}

	return NULL;
}

void PartitionNode::reparent(Partition& p)
{
	p.setParent(this);

	if (!isRoot())
		p.setRoles(PartitionRole(PartitionRole::Logical));
	else if (!p.roles().has(PartitionRole::Extended))
		p.setRoles(PartitionRole(PartitionRole::Primary));
	else
		p.setRoles(PartitionRole(PartitionRole::Extended));
}

qint32 PartitionNode::highestMountedChild() const
{
	qint32 result = -1;

	foreach (const Partition* p, children())
		if (p->number() > result && p->isMounted())
			result = p->number();
	
	return result;
}
