/***************************************************************************
 *   Copyright (C) 2008 by Volker Lanz <vl@fidra.de>                       *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA            *
 ***************************************************************************/

#if !defined(OPERATION__H)

#define OPERATION__H

#include <QObject>
#include <QList>
#include <qglobal.h>

class Partition;
class Device;
class OperationStack;
class Job;
class OperationRunner;
class Report;

class QString;
class QIcon;

// Rules for creating new operations that inherit the Operation class:
// (1) Don't modify anything in the ctor. The ctor runs before merging operations. If you
//     modify anything there, undo and merging will break. Just remember what you're
//     supposed to do in the ctor and perform modifications in preview().
// (2) Do not access the preview partitions and devices in description(). If you do,
//     the operation descriptions will be wrong.

class Operation : public QObject
{
	Q_OBJECT
	Q_DISABLE_COPY(Operation);

	friend class OperationStack;
	friend class OperationRunner;
	
	public:
		enum OperationStatus
		{
			StatusNone = 0,
			StatusPending,
			StatusRunning,
			StatusFinishedSuccess,
			StatusFinishedWarning,
			StatusError
		};

	protected:
		Operation();
		virtual ~Operation();

	signals:
		int progressChanged(int);
		void jobStarted(Job*, Operation*);
		void jobFinished(Job*, Operation*);
		
	public:
		virtual QString iconName() const = 0;
		virtual QString description() const = 0;
 		virtual void preview() = 0;
		virtual void undo() = 0;
		virtual bool execute(Report& parent);

		virtual OperationStatus status() const { return m_Status; }
		virtual QString statusText() const;
		virtual QIcon statusIcon() const;
		
		virtual void setStatus(OperationStatus s) { m_Status = s; }
		
		qint32 totalProgress() const;

	protected slots:
		void onJobStarted();
		void onJobFinished();
		
	protected:
		void insertPreviewPartition(Device& targetDevice, Partition& newPartition);
		void removePreviewPartition(Device& device, Partition& p);
		
		void addJob(Job* job);
		
		QList<Job*>& jobs() { return m_Jobs; }
		const QList<Job*>& jobs() const { return m_Jobs; }
		
		void setProgressBase(qint32 i) { m_ProgressBase = i; }
		qint32 progressBase() const { return m_ProgressBase; }
		
	private:
		OperationStatus m_Status;
		QList<Job*> m_Jobs;
		qint32 m_ProgressBase;
};

#endif
